﻿namespace HeuristicLab.Algorithms.PushGP.Expressions
{
  using HeuristicLab.Algorithms.PushGP.Interpreter;
  using HeuristicLab.Algorithms.PushGP.Stack;

  /// <summary>
  ///     Duplicates the top item on the INTEGER stack. Does not pop its argument
  ///     (which, if it did, would negate the effect of the duplication!).
  /// </summary>
  /// <typeparam name="T">Stacktype</typeparam>
  public abstract class DuplicateExpression<T> : StatelessExpression
  {
    protected void Eval(IStack<T> stack)
    {
      if (stack.Count == 0) return;

      stack.Push(stack.Top);
    }

    protected void Eval(IStack<Expression> stack)
    {
      if (stack.Count == 0) return;

      var top = stack.Top;

      stack.Push(top);
    }
  }

  public class IntegerDuplicateExpression : DuplicateExpression<long>
  {
    protected override string InitStringRepresentation()
    {
      return "INTEGER.DUP";
    }

    public override void Eval(IPushGpInterpreter interpreter)
    {
      this.Eval(interpreter.IntegerStack);
    }
  }

  public class FloatDuplicateExpression : DuplicateExpression<double>
  {
    protected override string InitStringRepresentation()
    {
      return "FLOAT.DUP";
    }

    public override void Eval(IPushGpInterpreter interpreter)
    {
      this.Eval(interpreter.FloatStack);
    }
  }

  public class BooleanDuplicateExpression : DuplicateExpression<bool>
  {
    protected override string InitStringRepresentation()
    {
      return "BOOLEAN.DUP";
    }

    public override void Eval(IPushGpInterpreter interpreter)
    {
      this.Eval(interpreter.BooleanStack);
    }
  }

  public class NameDuplicateExpression : DuplicateExpression<string>
  {
    protected override string InitStringRepresentation()
    {
      return "NAME.DUP";
    }

    public override void Eval(IPushGpInterpreter interpreter)
    {
      this.Eval(interpreter.NameStack);
    }
  }

  public class ExecDuplicateExpression : DuplicateExpression<Expression>
  {
    protected override string InitStringRepresentation()
    {
      return "EXEC.DUP";
    }

    public override void Eval(IPushGpInterpreter interpreter)
    {
      this.Eval(interpreter.ExecStack);
    }
  }

  public class CodeDuplicateExpression : DuplicateExpression<Expression>
  {
    protected override string InitStringRepresentation()
    {
      return "CODE.DUP";
    }

    public override void Eval(IPushGpInterpreter interpreter)
    {
      this.Eval(interpreter.CodeStack);
    }
  }
}