﻿namespace HeuristicLab.Algorithms.PushGP.Expressions {
  using System.Linq;
  using System.Threading;

  using HeuristicLab.Algorithms.PushGP.Data.Random;
  using HeuristicLab.Algorithms.PushGP.Interpreter;
  using HeuristicLab.Core;

  public class NameDefineXExecExpression : StatefullExpression<string> {

    public NameDefineXExecExpression(string state) : base(state) { }

    public override void Eval(IPushGpInterpreter interpreter) {
      Expression expression;
      if (!interpreter.IsNameQuoteFlagSet && interpreter.CustomExpressions.TryGetValue(this.State, out expression)) {
        interpreter.ExecStack.Push(expression);
      } else {
        interpreter.NameStack.Push(this.State);
        interpreter.IsNameQuoteFlagSet = false;
      }
    }

    public override bool Equals(object obj) {
      if (!base.Equals(obj)) return false;

      var other = obj as NameDefineXExecExpression;

      return this.State.Equals(other.State);
    }

    public override int GetHashCode() {
      return base.GetHashCode();
    }

    protected override string InitStringRepresentation() {
      return this.State;
    }

    protected override int InitId() {
      var hash = 19;

      hash = hash * 31 + this.GetType().GetHashCode();
      hash = hash * 31 + this.State.GetHashCode();

      return hash;
    }
  }

  /// <summary>
  ///     Sets a flag indicating that the next State encountered will be pushed onto the NAME stack (and not have its
  ///     associated value
  ///     pushed onto the EXEC stack), regardless of whether or not it has a definition. Upon encountering such a State and
  ///     pushing it
  ///     onto the NAME stack the flag will be cleared (whether or not the pushed State had a definition).
  /// </summary>
  public class NameQuoteExpression : StatelessExpression {
    protected override string InitStringRepresentation() {
      return "NAME.QUOTE";
    }

    public override void Eval(IPushGpInterpreter interpreter) {
      interpreter.IsNameQuoteFlagSet = true;
    }
  }

  /// <summary>
  ///     Pushes a randomly selected NAME that already has a definition.
  /// </summary>
  public class NameRandBoundNameExpression : StatelessExpression {
    private static readonly ThreadLocal<IRandom> Rand = RandomFactory.GetRandom();

    protected override string InitStringRepresentation() {
      return "NAME.RANDBOUNDNAME";
    }

    public override void Eval(IPushGpInterpreter interpreter) {
      if (!interpreter.CustomExpressions.Any()) return;

      var index = Rand.Value.Next(0, interpreter.CustomExpressions.Keys.Count - 1);
      var state = interpreter.CustomExpressions.Keys.ElementAt(index);

      interpreter.NameStack.Push(state);
    }
  }
}